# ################### Movetables describe the transformation of the coordinates by cube moves. #########################
import os
from os import path
import array as ar
from supercube.kociemba_2x2 import cubie as cb
from supercube.kociemba_2x2 import enums
from supercube.kociemba_2x2.defs import N_TWIST, N_CORNERS, N_MOVE

a = cb.CubieCube()
# ############################ Move table for the the corners. ##################################

# The twist coordinate describes the 3^6 = 729 possible orientations of the 8 corners
fname = ".move_corntwist"
#print("creating " + fname + " table...")
corntwist_move = ar.array('H', [0 for i in range(N_TWIST * N_MOVE)])
for i in range(N_TWIST):
    a.set_cornertwist(i)
    for j in (enums.Color.U, enums.Color.R, enums.Color.F):  # three faces U, R, F
        for k in range(3):  # three moves for each face, for example U, U2, U3 = U'
            a.multiply(cb.basicMoveCube[j])
            corntwist_move[N_MOVE * i + 3 * j + k] = a.get_corntwist()
        a.multiply(cb.basicMoveCube[j])  # 4. move restores face


# The corners coordinate describes the 7! = 5040 permutations of the corners.
fname = ".move_cornperm"
if not path.isfile(fname):
    #print("creating " + fname + " table...")
    cornperm_move = ar.array('H', [0 for i in range(N_CORNERS * N_MOVE)])
    # Move table for the corners. corner  < 40320
    for i in range(N_CORNERS):
        #if (i+1) % 200 == 0:
            #print('.', end='', flush=True)
        a.set_corners(i)
        for j in (enums.Color.U, enums.Color.R, enums.Color.F):  # three faces U, R, F
            for k in range(3):
                a.multiply(cb.basicMoveCube[j])
                cornperm_move[N_MOVE * i + 3 * j + k] = a.get_cornperm()
            a.multiply(cb.basicMoveCube[j])
    fh = open(fname, "wb")
    cornperm_move.tofile(fh)
    fh.close()
else:
    #print("loading " + fname + " table...")
    fh = open(fname, "rb")
    cornperm_move = ar.array('H')
    cornperm_move.fromfile(fh, N_CORNERS * N_MOVE)
fh.close()
########################################################################################################################



#  ####################  Enumerations which improve the readability of the code ########################################

from enum import IntEnum


class Facelet(IntEnum):
    """""
    The names of the facelet positions of the cube
              |********|
              |*U1**U2*|
              |********|
              |*U3**U4*|
              |********|
     |********|********|********|********|
     |*L1**L2*|*F1**F2*|*R1**R2*|*B1**B2*|
     |********|********|********|********|
     |*L3**L4*|*F3**F4*|*R3**R4*|*B3**B4*|
     |********|********|********|********|
              |********|
              |*D1**D2*|
              |********|
              |*D3**D4*|
              |********|
     

    A cube definition string "UBL..." means for example: In position U1 we have the U-color, in position U2 we have the
    B-color, in position U3 we have the L color etc. according to the order U1, U2, U3, U4, R1, R2, R3, R4, F1, F2, F3,
    F4, D1, D2, D3, D4, L1, L2, L3, L4, B1, B2, B3, B4 of the enum constants.
    """
    U1 = 0
    U2 = 1
    U3 = 2
    U4 = 3
    R1 = 4
    R2 = 5
    R3 = 6
    R4 = 7
    F1 = 8
    F2 = 9
    F3 = 10
    F4 = 11
    D1 = 12
    D2 = 13
    D3 = 14
    D4 = 15
    L1 = 16
    L2 = 17
    L3 = 18
    L4 = 19
    B1 = 20
    B2 = 21
    B3 = 22
    B4 = 23


class Color(IntEnum):
    """ The possible colors of the cube facelets. Color U refers to the color of the U(p)-face etc. 
    Also used to name the faces itself."""
    U = 0
    R = 1
    F = 2
    D = 3
    L = 4
    B = 5


class Corner(IntEnum):
    """The names of the corner positions of the cube. Corner URF e.g. has an U(p), a R(ight) and a F(ront) facelet."""
    URF = 0
    UFL = 1
    ULB = 2
    UBR = 3
    DRB = 4
    DFR = 5
    DLF = 6
    DBL = 7


class Move(IntEnum):
    """The moves in the faceturn metric. Not to be confused with the names of the facelet positions in class Facelet."""
    U1 = 0
    U2 = 1
    U3 = 2
    R1 = 3
    R2 = 4
    R3 = 5
    F1 = 6
    F2 = 7
    F3 = 8



# ################## The code of the server socket which communicates with the client ##################################

import socket
import sys
import threading
import solver
import time


def client_thread(conn):
    while True:  # infinite loop only necessary for telnet client
        # Receiving from client
        data = []
        while not (ord('\n') in data or ord('\r') in data):
            try:
                a = conn.recv(1024).upper()
                if len(a) == 0:
                    conn.close()
                    print('Connection closed', flush=True)
                    return
            except:
                print('Connection closed', flush=True)
                conn.close()
                return
            for i in range(len(a)):
                if a[i] in [ord('\n'), ord('\r'), ord('G'), ord('E'), ord('T'),  ord('U'), ord('R'), ord('F'), ord('D'),
                            ord('L'), ord('B')]:
                    data.append(a[i])
        if data[0] == ord('X'):
            break
        defstr = ''.join(chr(i) for i in data if chr(i) > chr(32))
        qpos = defstr.find('GET')
        if qpos >= 0:  # in this case we suppose the client is a webbrowser
            defstr = defstr[qpos+3:qpos+27]
            reply = 'HTTP/1.1 200 OK' + '\n\n' + '<html><head><title>Answer from Cubesolver</title></head><body>' + '\n'
            maneuverlist = solver.solve(defstr).split('\r\n')
            for m in maneuverlist:
                reply += m +'<br>\n'
            reply += '</body></html>\n'
            conn.sendall(reply.encode())
            conn.close()
        else:  # other client, for example the GUI client or telnet
            reply = (solver.solve(defstr)+'\n').encode()
            print(defstr)
            try:
                conn.sendall(reply)
            except:
                print('Error while sending data. Connection closed', flush=True)
                conn.close()
                return
    conn.close()


def server_start(args):
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    print('Server socket created')
    try:
        s.bind(('', int(args[1])))  # bind socket to local host and port
    except socket.error as e:
        print('Server socket bind failed. Error Code : ' + str(e.errno))
        sys.exit()
    s.listen(10)
    print('Server now listening...')

    while 1:
        conn, addr = s.accept()
        print('Connected with ' + addr[0] + ':' + str(addr[1]) + ', ' + time.strftime("%Y.%m.%d  %H:%M:%S"))
        threading.Thread(target=client_thread, args=(conn,)).start()
    s.close()




import sys
from kociemba import solve


def main():
    if len(sys.argv) == 2:
        print(solve(sys.argv[1]))
    elif len(sys.argv) == 3:
        print(solve(sys.argv[1], sys.argv[2]))
    else:
        print('Usage: kociemba <cubestring> [resulting_cubestring]\nfor example:\nkociemba DRLUUBFBRBLURRLRUBLRDDFDLFUFUFFDBRDUBRUFLLFDDBFLUBLRBD BBURUDBFUFFFRRFUUFLULUFUDLRRDBBDBDBLUDDFLLRRBRLLLBRDDF')



# ##################### The pruning table cuts the search tree during the search. ######################################
# ########################## In this case it it gives the exact distance to the solved state. ##########################

from supercube.kociemba_2x2 import defs
from supercube.kociemba_2x2 import enums
from supercube.kociemba_2x2 import moves as mv
from os import path
import os
import array as ar

corner_depth = None


def create_cornerprun_table():
    """Creates/loads the corner pruning table."""
    fname = ".cornerprun"
    global corner_depth
    if not path.isfile(fname):
        #print("creating " + fname + " table...")
        corner_depth = ar.array('b', [-1] * (defs.N_CORNERS * defs.N_TWIST))
        corners = 0  # values for solved cube
        twist = 0
        corner_depth[defs.N_TWIST * corners + twist] = 0
        done = 1
        depth = 0
        while done != defs.N_CORNERS * defs.N_TWIST:
            for corners in range(defs.N_CORNERS):
                for twist in range(defs.N_TWIST):
                    if corner_depth[defs.N_TWIST * corners + twist] == depth:
                        for m in enums.Move:
                            corners1 = mv.cornperm_move[9 * corners + m]
                            twist1 = mv.corntwist_move[9 * twist + m]
                            idx1 = defs.N_TWIST * corners1 + twist1
                            if corner_depth[idx1] == -1:  # entry not yet filled
                                corner_depth[idx1] = depth + 1
                                done += 1

            depth += 1
        fh = open(fname, "wb")
        corner_depth.tofile(fh)
    else:
        #print("loading " + fname + " table...")
        fh = open(fname, "rb")
        corner_depth = ar.array('b')
        corner_depth.fromfile(fh, defs.N_CORNERS * defs.N_TWIST)
    fh.close()

create_cornerprun_table()




# ################# Start the server and listen for connections ########################################################

import sockets
import sys


if __name__ == '__main__':  # file is executed
    if len(sys.argv) < 2:
        sys.argv.append(str(8080))  # Port 8080 default port
    print('startserver')
    sockets.server_start(sys.argv)
else:
    def start(port):
        sockets.server_start((-1, port))



from supercube.cube import Cube
from supercube.kociemba_2x2 import solver
import os

class PocketCube(Cube):
    _SIZE_OF_CUBE = 2

    def solve(self):
        kociemba_state = ""

        for i in range(2):
            for j in range(2):
                kociemba_state += self._get_face_by_element(self._STATE_OF_CUBE[0][i][j])
        for i in range(2):
            for j in range(2):
                kociemba_state += self._get_face_by_element(self._STATE_OF_CUBE[3][i][j])
        for i in range(2):
            for j in range(2):
                kociemba_state += self._get_face_by_element(self._STATE_OF_CUBE[2][i][j])
        for i in range(2):
            for j in range(2):
                kociemba_state += self._get_face_by_element(self._STATE_OF_CUBE[5][i][j])
        for i in range(2):
            for j in range(2):
                kociemba_state += self._get_face_by_element(self._STATE_OF_CUBE[1][i][j])
        for i in range(2):
            for j in range(2):
                kociemba_state += self._get_face_by_element(self._STATE_OF_CUBE[4][i][j])
        kociemba_moves = solver.solve(kociemba_state)
        self.move(kociemba_moves)
        return kociemba_moves


from supercube.cube import Cube


class NCube(Cube):

    def _SIZE_OF_CUBE(self) -> int:
        pass

    def solve(self):
        raise Exception("N Cube cannot be solved!")

    def __init__(self, size_of_cube):
        self._SIZE_OF_CUBE = size_of_cube
        if (self._SIZE_OF_CUBE < 2):
            raise Exception("Size of cube cannot be less than 2")
        super().__init__()



# ###################### The solve function computes all optimal solving maneuvers #####################################
from supercube.kociemba_2x2 import face
from supercube.kociemba_2x2 import cubie
from supercube.kociemba_2x2 import coord
from supercube.kociemba_2x2 import enums as en
from supercube.kociemba_2x2 import moves as mv
from supercube.kociemba_2x2 import pruning as pr
from supercube.kociemba_2x2.defs import N_TWIST

solutions = []


def search(cornperm, corntwist, sofar, togo):
    global solutions
    # ##############################################################################################################
    if togo == 0:
        if len(solutions) == 0 or (len(solutions[-1]) == len(sofar)):
            solutions.append(sofar[:])
    else:
        for m in en.Move:

            if len(sofar) > 0:
                if sofar[-1] // 3 == m // 3: # successive moves on same face
                    continue

            cornperm_new = mv.cornperm_move[9 * cornperm + m]
            corntwist_new = mv.corntwist_move[9 * corntwist + m]

            if pr.corner_depth[N_TWIST * cornperm_new + corntwist_new] >= togo:
                continue  # impossible to reach solved cube in togo - 1 moves

            sofar.append(m)
            search(cornperm_new, corntwist_new, sofar, togo - 1)
            sofar.pop(-1)


def solve(cubestring):
    """Solves a 2x2x2 cube defined by its cube definition string.
     :param cubestring: The format of the string is given in the Facelet class defined in the file enums.py
     :return A list of all optimal solving maneuvers
    """
    global solutions
    fc = face.FaceCube()
    s = fc.from_string(cubestring) #####################################################################################
    if s is not True:
        return s  # Error in facelet cube
    cc = fc.to_cubie_cube()
    s = cc.verify()
    if s != cubie.CUBE_OK:
        return s  # Error in cubie cube

    solutions = []
    co_cube = coord.CoordCube(cc)

    togo = pr.corner_depth[N_TWIST * co_cube.cornperm + co_cube.corntwist]
    search(co_cube.cornperm, co_cube.corntwist, [], togo)

    s = ''
    for i in range(min(len(solutions), 1)):  # use  range(min(len(solutions), 1)) if you want to return only a single solution
        ps = ''
        for m in solutions[i]:
            ps += m.name + ' '
        s += ps
    new_s = ''
    for move in s.strip().split(" "):
        if move[-1] == "1":
            new_s += move[0] + " "
        elif move[-1] == "2":
            new_s += move[0] + "2 "
        elif move[-1] == "3":
            new_s += move[0] + "' "
    return new_s
########################################################################################################################




# ####### The 2x2x2 cube on the facelet level is described by positions of the colored stickers. #######################

from supercube.kociemba_2x2.defs import cornerFacelet, cornerColor
from supercube.kociemba_2x2.enums import Color, Corner
from supercube.kociemba_2x2.cubie import CubieCube, CUBE_OK


class FaceCube:
    """Represents a 2x2x2 cube on the facelet level with 24 colored facelets."""
    def __init__(self):
        self.f = []
        for i in range(4):
            self.f.append(Color.U)
        for i in range(4):
            self.f.append(Color.R)
        for i in range(4):
            self.f.append(Color.F)
        for i in range(4):
            self.f.append(Color.D)
        for i in range(4):
            self.f.append(Color.L)
        for i in range(4):
            self.f.append(Color.B)

    def __str__(self):
        return self.to_string()

    def from_string(self, s):
        """Constructs a facelet cube from a string. See class Facelet(IntEnum) in enums.py for string format.
        
        The color scheme is detected automatically.
        """
        if len(s) < 24:
            return 'Error: Cube definition string ' + s + ' contains less than 24 facelets.'
        elif len(s) > 24:
            return 'Error: Cube definition string ' + s + ' contains more than 24 facelets.'
        cnt = [0] * 6
        for i in range(24):
            if s[i] == 'U':
                self.f[i] = Color.U
                cnt[Color.U] += 1
            elif s[i] == 'R':
                self.f[i] = Color.R
                cnt[Color.R] += 1
            elif s[i] == 'F':
                self.f[i] = Color.F
                cnt[Color.F] += 1
            elif s[i] == 'D':
                self.f[i] = Color.D
                cnt[Color.D] += 1
            elif s[i] == 'L':
                self.f[i] = Color.L
                cnt[Color.L] += 1
            elif s[i] == 'B':
                self.f[i] = Color.B
                cnt[Color.B] += 1
        if all(x == 4 for x in cnt):
            # remap colors if necessary
            col = [self.f[cornerFacelet[Corner.DBL][i]] for i in range(3)]  # colors of the DBL-corner
            map_col = [-1] * 6
            for i in range(3):
                map_col[col[i]] = cornerColor[Corner.DBL][i]  # map colors to right colors
            # now remap the remaining colors, try all possibilites
            a = ((Color.U, Color.R, Color.F), (Color.U, Color.F, Color.R), (Color.R, Color.U, Color.F),
                 (Color.R, Color.F, Color.U), (Color.F, Color.U, Color.R), (Color.F, Color.R, Color.U))
            empty = []
            for i in Color:
                if map_col[i] == -1:
                    empty.append(i)  # empty contains the 3 indices of the yet nonmapped colors
            fsave = self.f[:]
            for c in a:
                for i in range(3):
                    map_col[empty[i]] = c[i]
                self.f = fsave[:]
                for i in range(24):
                    self.f[i] = map_col[self.f[i]]  # remap the colors
                cc = self.to_cubie_cube()
                s = cc.verify()
                if s == CUBE_OK:
                    return True

            return 'Error: Facelet configuration does not define a valid cube.'
        else:
            return 'Error: Cube definition string ' + s + ' does not contain exactly 4 facelets of each color.'



    def to_string(self):
        """Gives string representation of the facelet cube."""
        s = ''
        for i in range(24):
            if self.f[i] == Color.U:
                s += 'U'
            elif self.f[i] == Color.R:
                s += 'R'
            elif self.f[i] == Color.F:
                s += 'F'
            elif self.f[i] == Color.D:
                s += 'D'
            elif self.f[i] == Color.L:
                s += 'L'
            elif self.f[i] == Color.B:
                s += 'B'
        return s

    def to_cubie_cube(self):
        """Returns a cubie representation of the facelet cube."""
        cc = CubieCube()
        cc.cp = [-1] * 8  # invalidate corner permutation
        for i in Corner:
            col0 = None
            fac = cornerFacelet[i]  # facelets of corner  at position i
            for ori in range(3):
                if self.f[fac[ori]] == Color.U or self.f[fac[ori]] == Color.D:
                    col0 = self.f[fac[ori]]
                    break
            col1 = self.f[fac[(ori + 1) % 3]]  # colors which identify the corner at position i
            col2 = self.f[fac[(ori + 2) % 3]]
            for j in Corner:
                col = cornerColor[j]  # colors of corner j
                if col0 != None and col0 == col[0] and col1 == col[1] and col2 == col[2]:
                    cc.cp[i] = j  # we have corner j in corner position i
                    cc.co[i] = ori
                    break
        return cc



# ################ A simple graphical interface which communicates with the server #####################################

from tkinter import *
import socket
import face
import cubie

# ################################ Edit the following line to use different colors #####################################
cols = ("yellow", "green", "red", "white", "blue", "orange")
########################################################################################################################

# ################################## some global variables and constants ###############################################
DEFAULT_HOST = 'localhost'
DEFAULT_PORT = '8080'
width = 60  # width of a facelet in pixels
facelet_id = [[[0 for col in range(2)] for row in range(2)] for face in range(6)]
colorpick_id = [0 for i in range(6)]
curcol = None
t = ("U", "R", "F", "D", "L", "B")
########################################################################################################################

# ################################################ Diverse functions ###################################################


def show_text(txt):
    """Displays messages."""
    print(txt)
    display.insert(END, txt)
    root.update_idletasks()


def create_facelet_rects(a):
    """Initializes the facelet grid on the canvas"""
    offset = ((1, 0), (2, 1), (1, 1), (1, 2), (0, 1), (3, 1))
    for f in range(6):
        for row in range(2):
            y = 20 + offset[f][1] * 2 * a + row * a
            for col in range(2):
                x = 20 + offset[f][0] * 2 * a + col * a
                facelet_id[f][row][col] = canvas.create_rectangle(x, y, x + a, y + a, fill="grey")
                if row == 1 and col == 1:
                    canvas.create_text(x, y, font=("", 14), text=t[f], state=DISABLED)


def create_colorpick_rects(a):
    """Initializes the "paintbox" on the canvas"""
    global curcol
    global cols
    for i in range(6):
        x = (i % 3)*(a+5) + 5*a
        y = (i // 3)*(a+5) + 5*a
        colorpick_id[i] = canvas.create_rectangle(x, y, x + a, y + a, fill=cols[i])
        canvas.itemconfig(colorpick_id[0], width=4)
        curcol = cols[0]


def get_definition_string():
    """Generates the cube definition string from the facelet colors."""
    color_to_facelet = {}
    for i in range(6):
        color_to_facelet.update({cols[i]: t[i]})
    s = ''
    for f in range(6):
        for row in range(2):
            for col in range(2):
                s += color_to_facelet[canvas.itemcget(facelet_id[f][row][col], "fill")]
    return s
########################################################################################################################

# ############################### Solve the displayed cube with a local or remote server ###############################


def solve():
    """Connects to the server and returns the solving maneuver"""
    display.delete(1.0, END)  # clear output window
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    except socket.error:
        show_text('Failed to create socket')
        return
    # host = 'f9f0b2jt6zmzyo6b.myfritz.net'  # my RaspberryPi, if online
    host = txt_host.get(1.0, END).rstrip()  # default is localhost
    port = int(txt_port.get(1.0, END))  # default is port 8080

    try:
        remote_ip = socket.gethostbyname(host)
    except socket.gaierror:
        show_text('Hostname could not be resolved.')
        return
    try:
        s.connect((remote_ip, port))
    except:
        show_text('Cannot connect to server!')
        return
    show_text('Connected with ' + remote_ip + '\n')
    try:
        defstr = get_definition_string()+'\n'
    except:
        show_text('Invalid facelet configuration.\nWrong or missing colors.')
        return
    # show_text(defstr)
    try:
        s.sendall((defstr+'\n').encode())
    except:
        show_text('Cannot send cube configuration to server.')
        return
    show_text(s.recv(10000).decode())  # big buffer because we return all solutions (196 solutions in "worst" case)
########################################################################################################################

# ################################# Functions to change the facelet colors #############################################


def clean():
    """Restores the cube to a clean cube."""
    for f in range(6):
        for row in range(2):
            for col in range(2):
                canvas.itemconfig(facelet_id[f][row][col], fill=cols[f])


def empty():
    """Removes the facelet colors."""
    for f in range(6):
        for row in range(2):
            for col in range(2):
                canvas.itemconfig(facelet_id[f][row][col], fill="grey")


def random():
    """Generates a random cube and sets the corresponding facelet colors."""
    cc = cubie.CubieCube()
    cc.randomize()
    fc = cc.to_facelet_cube()
    idx = 0
    for f in range(6):
        for row in range(2):
            for col in range(2):
                canvas.itemconfig(facelet_id[f][row][col], fill=cols[fc.f[idx]] )
                idx += 1
########################################################################################################################

# ################################### Edit the facelet colors ##########################################################


def click(event):
    """Defines how to react on left mouseclicks."""
    global curcol
    idlist = canvas.find_withtag("current")
    if len(idlist) > 0:
        if idlist[0] in colorpick_id:
            curcol = canvas.itemcget("current", "fill")
            for i in range(6):
                canvas.itemconfig(colorpick_id[i], width=1)
            canvas.itemconfig("current", width=5)
        else:
            canvas.itemconfig("current", fill=curcol)
########################################################################################################################

#  ###################################### Generate and display the TK_widgets ##########################################
root = Tk()
root.wm_title("Solver Client")
canvas = Canvas(root, width=10 * width + 20, height=7.5 * width + 20)
canvas.pack()

bsolve = Button(text="Solve", height=2, width=10, relief=RAISED, command=solve)
bsolve_window = canvas.create_window(10 + 8.5 * width, 10 + 4.5 * width, anchor=NW, window=bsolve)
bclean = Button(text="Clean", height=1, width=10, relief=RAISED, command=clean)
bclean_window = canvas.create_window(10 + 8.5 * width, 10 + 5.5 * width, anchor=NW, window=bclean)
bempty = Button(text="Empty", height=1, width=10, relief=RAISED, command=empty)
bempty_window = canvas.create_window(10 + 8.5 * width, 10 + 6 * width, anchor=NW, window=bempty)
brandom = Button(text="Random", height=1, width=10, relief=RAISED, command=random)
brandom_window = canvas.create_window(10 + 8.5 * width, 10 + 6.5 * width, anchor=NW, window=brandom)
display = Text(height=7, width=39)
text_window = canvas.create_window(10 + 4.5 * width, 10 + 0 * width, anchor=NW, window=display)
hp = Label(text='    Hostname and Port')
hp_window = canvas.create_window(10 + 0 * width, 10 + 6.1 * width, anchor=NW, window=hp)
txt_host = Text(height=1, width=20)
txt_host_window = canvas.create_window(10 + 0 * width, 10 + 6.5 * width, anchor=NW, window=txt_host)
txt_host.insert(INSERT, DEFAULT_HOST)
txt_port = Text(height=1, width=20)
txt_port_window = canvas.create_window(10 + 0 * width, 10 + 7 * width, anchor=NW, window=txt_port)
txt_port.insert(INSERT, DEFAULT_PORT)
canvas.bind("<Button-1>", click)
create_facelet_rects(width)
create_colorpick_rects(width)
clean()
root.mainloop()
########################################################################################################################




# ######### The 2x2x2 cube on the cubie level is described by the permutation and orientations of the corners ##########

from supercube.kociemba_2x2.defs import cornerFacelet, cornerColor, N_CORNERS, N_TWIST
from supercube.kociemba_2x2.enums import Color, Corner as Co
from supercube.kociemba_2x2 import face
from supercube.kociemba_2x2.misc import rotate_left, rotate_right
from random import randrange

# ################## the basic six cube moves described by permutations and changes in orientation #####################

# Up-move
cpU = [Co.UBR, Co.URF, Co.UFL, Co.ULB,  Co.DRB, Co.DFR, Co.DLF, Co.DBL]
coU = [0, 0, 0, 0, 0, 0, 0, 0]

# Right-move
cpR = [Co.DFR, Co.UFL, Co.ULB, Co.URF, Co.UBR, Co.DRB, Co.DLF, Co.DBL]  # permutation of the corners
coR = [2, 0, 0, 1, 2, 1, 0, 0]  # changes of the orientations of the corners

# Front-move
cpF = [Co.UFL, Co.DLF, Co.ULB, Co.UBR, Co.DRB, Co.URF, Co.DFR, Co.DBL]
coF = [1, 2, 0, 0, 0, 2, 1, 0]

########################################################################################################################

CUBE_OK = True


class CubieCube:
    """Represents a 2x2x2 cube on the cubie level with 8 corner cubies and the corner orientations.

    Is also used to represent the 18 cube moves.
    """

    def __init__(self, cp=None, co=None):
        """
        Initializes corners and edges.
        :param cp: corner permutation
        :param co: corner orientation
        """
        if cp is None:
            self.cp = [Co(i) for i in range(8)]  # You may not put this as the default two lines above!
        else:
            self.cp = cp[:]
        if co is None:
            self.co = [0] * 8
        else:
            self.co = co[:]

    def __str__(self):
        """Prints string for a cubie cube."""
        s = ''
        for i in Co:
            s = s + '(' + str(self.cp[i]) + ',' + str(self.co[i]) + ')'
        return s

    def __eq__(self, other):
        """Defines equality of two cubie cubes."""
        if self.cp == other.cp and self.co == other.co:
            return True
        else:
            return False

    def to_facelet_cube(self):
        """Returns a facelet representation of the cube."""
        fc = face.FaceCube()
        for i in Co:
            j = self.cp[i]  # corner j is at corner position i
            ori = self.co[i]  # orientation of C j at position i
            for k in range(3):
                fc.f[cornerFacelet[i][(k + ori) % 3]] = cornerColor[j][k]
        return fc

    def multiply(self, b):
        """Multiplies this cubie cube with another cubie cube b. Does not change b."""
        c_perm = [0] * 8
        c_ori = [0] * 8
        ori = 0
        for c in Co:
            c_perm[c] = self.cp[b.cp[c]]
            ori_a = self.co[b.cp[c]]
            ori_b = b.co[c]
            if ori_a < 3 and ori_b < 3:  # two regular cubes
                ori = ori_a + ori_b
                if ori >= 3:
                    ori -= 3
            elif ori_a < 3 <= ori_b:  # cube b is in a mirrored state
                ori = ori_a + ori_b
                if ori >= 6:
                    ori -= 3  # the composition also is in a mirrored state
            elif ori_a >= 3 > ori_b:  # cube a is in a mirrored state
                ori = ori_a - ori_b
                if ori < 3:
                    ori += 3  # the composition is a mirrored cube
            elif ori_a >= 3 and ori_b >= 3:  # if both cubes are in mirrored states
                ori = ori_a - ori_b
                if ori < 0:
                    ori += 3  # the composition is a regular cube
            c_ori[c] = ori
        for c in Co:
            self.cp[c] = c_perm[c]
            self.co[c] = c_ori[c]

    def inv_cubie_cube(self, d):
        """Stores the inverse of this cubie cube in d."""

        for c in Co:
            d.cp[self.cp[c]] = c
        for c in Co:
            ori = self.co[d.cp[c]]
            if ori >= 3:
                d.co[c] = ori
            else:
                d.co[c] = -ori
                if d.co[c] < 0:
                    d.co[c] += 3


    # ###################################### coordinates for 2x2x2 cube #################################################
    def get_corntwist(self):
        """The twist of the 8 corners. 0 <= twist < 729. The DBL-corner is fixed."""
        ret = 0
        for i in range(Co.URF, Co.DLF):
            ret = 3 * ret + self.co[i]
        return ret

    def set_cornertwist(self, twist):
        twistparity = 0
        for i in range(Co.DLF - 1, Co.URF - 1, -1):
            self.co[i] = twist % 3
            twistparity += self.co[i]
            twist //= 3
        self.co[Co.DLF] = ((3 - twistparity % 3) % 3)


    def get_cornperm(self):
        """The permutation of the 8 corners. 0 <= corners < 5040. The DLB_corner is fixed."""
        perm = list(self.cp)  # duplicate cp
        b = 0
        for j in range(Co.DBL, Co.URF, -1):
            k = 0
            while perm[j] != j:
                rotate_left(perm, 0, j)
                k += 1
            b = (j + 1) * b + k
        return b

    def set_corners(self, idx):

        self.cp = [i for i in Co]
        for j in Co:
            k = idx % (j + 1)
            idx //= j + 1
            while k > 0:
                rotate_right(self.cp, 0, j)
                k -= 1
                # ############################## end coordinates for 2x2x2 cube ########################################

                # #################################### other usefull functions #########################################

    def randomize(self):
        """Generates a random cube. The probability is the same for all possible states."""
        self.set_corners(randrange(N_CORNERS))
        self.set_cornertwist(randrange(N_TWIST))

    def verify(self):
        """Checks if cubiecube is valid"""

        corner_count = [0] * 8
        for i in Co:
            corner_count[self.cp[i]] += 1
        for i in Co:
            if corner_count[i] != 1:
                return 'Error: Some corners are undefined.'

        s = 0
        for i in Co:
            s += self.co[i]
        if s % 3 != 0:
            return 'Error: Total corner twist is wrong.'

        return CUBE_OK


########################################################################################################################

# ################################## these cubes represent the basic cube moves ########################################
basicMoveCube = [0] * 6
basicMoveCube[Color.U] = CubieCube(cpU, coU)
basicMoveCube[Color.R] = CubieCube(cpR, coR)
basicMoveCube[Color.F] = CubieCube(cpF, coF)
########################################################################################################################

# ################################# these cubes represent the all 9 cube moves ########################################
moveCube = [0] * 9
for c1 in [Color.U, Color.R, Color.F]:
    cc = CubieCube()
    for k1 in range(3):
        cc.multiply(basicMoveCube[c1])
        moveCube[3 * c1 + k1] = CubieCube(cc.cp, cc.co)
########################################################################################################################



from abc import ABC, abstractmethod
from copy import deepcopy
import random
import re
import pyvista as pv
import numpy as np
import IPython


class Cube(ABC):
    @property
    @abstractmethod
    def _SIZE_OF_CUBE(self) -> int:
        pass

    @abstractmethod
    def solve(self):
        pass

    def __init__(self):
        self._UP_COLOR_ELEMENTS = []
        self._LEFT_COLOR_ELEMENTS = []
        self._FRONT_COLOR_ELEMENTS = []
        self._RIGHT_COLOR_ELEMENTS = []
        self._BACK_COLOR_ELEMENTS = []
        self._DOWN_COLOR_ELEMENTS = []
        self._STATE_OF_CUBE = self.__generate_clear_state()
        self._MAX_NUM_LEN = len(str(self._STATE_OF_CUBE[-1][-1][-1]))
        self._SQUARE_LEN = 2 + self._MAX_NUM_LEN * self._SIZE_OF_CUBE + 2 * (self._SIZE_OF_CUBE - 1)
        self._MAX_LAYER = self._SIZE_OF_CUBE
        self.GAP_SIZE = 0.01
        self.CENTER_3DCUBE_SIZE = (self._SIZE_OF_CUBE - 1) + self.GAP_SIZE * (self._SIZE_OF_CUBE - 1)

    def get_color_by_element(self, element: int):
        if element in self._UP_COLOR_ELEMENTS:
            return "W"
        elif element in self._BACK_COLOR_ELEMENTS:
            return "Y"
        elif element in self._LEFT_COLOR_ELEMENTS:
            return "O"
        elif element in self._RIGHT_COLOR_ELEMENTS:
            return "R"
        elif element in self._BACK_COLOR_ELEMENTS:
            return "B"
        elif element in self._FRONT_COLOR_ELEMENTS:
            return "G"
        else:
            raise Exception("No such element!")

    def __generate_clear_state(self):
        n = 1
        result = []
        total_number_of_elements = 6 * self._SIZE_OF_CUBE ** 2
        for face in range(6):
            result.append([])
            for row in range(self._SIZE_OF_CUBE):
                result[face].append([])
                for col in range(self._SIZE_OF_CUBE):
                    result[face][row].append(n)
                    n += 1
        sq = self._SIZE_OF_CUBE**2
        for element in range(1, total_number_of_elements+1):
            if element in range(1, sq+1):
                self._UP_COLOR_ELEMENTS.append(element)
            elif element in range(sq+1, sq*2+1):
                self._LEFT_COLOR_ELEMENTS.append(element)
            elif element in range(sq*2+1, sq*3+1):
                self._FRONT_COLOR_ELEMENTS.append(element)
            elif element in range(sq*3+1, sq*4+1):
                self._RIGHT_COLOR_ELEMENTS.append(element)
            elif element in range(sq*4+1, sq*5+1):
                self._BACK_COLOR_ELEMENTS.append(element)
            elif element in range(sq*5+1, sq*6+1):
                self._DOWN_COLOR_ELEMENTS.append(element)
        return result

    def f(self, clockwise=True, layer=1):
        if layer > self._MAX_LAYER or layer <= 0:
            raise Exception("Invalid layer!")
        if clockwise and layer == 1:
            self.__f_move_clockwise_external()
        elif clockwise and layer == self._SIZE_OF_CUBE:
            self.__b_move_anticlockwise_external()
        elif not clockwise and layer == self._SIZE_OF_CUBE:
            self.__b_move_clockwise_external()
        elif clockwise and layer > 1:
            self.__f_move_clockwise_internal(layer=layer)
        elif not clockwise and layer == 1:
            self.__f_move_anticlockwise_external()
        elif not clockwise and layer > 1:
            self.__f_move_anticlockwise_internal(layer=layer)

    def __clockwise_face_move(self, face_index, ORIGINAL_STATE):
        new_face = [[] for _ in range(self._SIZE_OF_CUBE)]
        row = 0
        for i in range(self._SIZE_OF_CUBE):
            for j in range(self._SIZE_OF_CUBE):
                new_face[row].append(ORIGINAL_STATE[face_index][self._SIZE_OF_CUBE - j - 1][i])
            row += 1
        self._STATE_OF_CUBE[face_index] = new_face

    def __anticlockwise_face_move(self, face_index, ORIGINAL_STATE):
        new_face = [[] for _ in range(self._SIZE_OF_CUBE)]
        row = 0
        for i in range(self._SIZE_OF_CUBE):
            for j in range(self._SIZE_OF_CUBE):
                new_face[row].append(ORIGINAL_STATE[face_index][j][self._SIZE_OF_CUBE - i - 1])
            row += 1
        self._STATE_OF_CUBE[face_index] = new_face

    def __f_move_clockwise_external(self):
        ORIGINAL_STATE = deepcopy(self._STATE_OF_CUBE)
        self.__clockwise_face_move(2, ORIGINAL_STATE)
        for i in range(self._SIZE_OF_CUBE):
            self._STATE_OF_CUBE[3][i][0] = ORIGINAL_STATE[0][-1][i]
            self._STATE_OF_CUBE[0][-1][i] = ORIGINAL_STATE[1][self._SIZE_OF_CUBE - i - 1][-1]
            self._STATE_OF_CUBE[5][0][i] = ORIGINAL_STATE[3][self._SIZE_OF_CUBE - i - 1][0]
            self._STATE_OF_CUBE[1][i][-1] = ORIGINAL_STATE[5][0][i]

    def __f_move_anticlockwise_external(self):
        ORIGINAL_STATE = deepcopy(self._STATE_OF_CUBE)
        self.__anticlockwise_face_move(2, ORIGINAL_STATE)
        for i in range(self._SIZE_OF_CUBE):
            self._STATE_OF_CUBE[0][-1][i] = ORIGINAL_STATE[3][i][0]
            self._STATE_OF_CUBE[3][i][0] = ORIGINAL_STATE[5][0][self._SIZE_OF_CUBE - i - 1]
            self._STATE_OF_CUBE[5][0][i] = ORIGINAL_STATE[1][i][-1]
            self._STATE_OF_CUBE[1][i][-1] = ORIGINAL_STATE[0][-1][self._SIZE_OF_CUBE - i - 1]

    def __f_move_clockwise_internal(self, layer):
        layer -= 1
        ORIGINAL_STATE = deepcopy(self._STATE_OF_CUBE)
        for i in range(self._SIZE_OF_CUBE):
            self._STATE_OF_CUBE[0][-1 - layer][i] = ORIGINAL_STATE[1][self._SIZE_OF_CUBE - i - 1][-1 - layer]
            self._STATE_OF_CUBE[1][i][-1 - layer] = ORIGINAL_STATE[5][layer][i]
            self._STATE_OF_CUBE[3][i][layer] = ORIGINAL_STATE[0][-1 - layer][i]
            self._STATE_OF_CUBE[5][layer][i] = ORIGINAL_STATE[3][self._SIZE_OF_CUBE - i - 1][layer]

    def __f_move_anticlockwise_internal(self, layer):
        layer -= 1
        ORIGINAL_STATE = deepcopy(self._STATE_OF_CUBE)
        for i in range(self._SIZE_OF_CUBE):
            self._STATE_OF_CUBE[0][-1 - layer][i] = ORIGINAL_STATE[3][i][layer]
            self._STATE_OF_CUBE[1][i][-1 - layer] = ORIGINAL_STATE[0][-1 - layer][self._SIZE_OF_CUBE - i - 1]
            self._STATE_OF_CUBE[3][i][layer] = ORIGINAL_STATE[5][layer][self._SIZE_OF_CUBE - i - 1]
            self._STATE_OF_CUBE[5][layer][i] = ORIGINAL_STATE[1][i][-1 - layer]

    def u(self, clockwise=True, layer=1):
        if layer > self._MAX_LAYER or layer <= 0:
            raise Exception("Invalid layer!")
        if clockwise and layer == 1:
            self.__u_move_clockwise_external()
        elif clockwise and layer == self._SIZE_OF_CUBE:
            self.__d_move_anticlockwise_external()
        elif not clockwise and layer == self._SIZE_OF_CUBE:
            self.__d_move_clockwise_external()
        elif clockwise and layer > 1:
            self.__u_move_clockwise_internal(layer=layer)
        elif not clockwise and layer == 1:
            self.__u_move_anticlockwise_external()
        elif not clockwise and layer > 1:
            self.__u_move_anticlockwise_internal(layer=layer)

    def __u_move_clockwise_external(self):
        ORIGINAL_STATE = deepcopy(self._STATE_OF_CUBE)
        self.__clockwise_face_move(0, ORIGINAL_STATE)
        for i in range(self._SIZE_OF_CUBE):
            self._STATE_OF_CUBE[1][0][i] = ORIGINAL_STATE[2][0][i]
            self._STATE_OF_CUBE[2][0][i] = ORIGINAL_STATE[3][0][i]
            self._STATE_OF_CUBE[3][0][i] = ORIGINAL_STATE[4][0][i]
            self._STATE_OF_CUBE[4][0][i] = ORIGINAL_STATE[1][0][i]

    def __u_move_anticlockwise_external(self):
        ORIGINAL_STATE = deepcopy(self._STATE_OF_CUBE)
        self.__anticlockwise_face_move(0, ORIGINAL_STATE)
        for i in range(self._SIZE_OF_CUBE):
            self._STATE_OF_CUBE[1][0][i] = ORIGINAL_STATE[4][0][i]
            self._STATE_OF_CUBE[2][0][i] = ORIGINAL_STATE[1][0][i]
            self._STATE_OF_CUBE[3][0][i] = ORIGINAL_STATE[2][0][i]
            self._STATE_OF_CUBE[4][0][i] = ORIGINAL_STATE[3][0][i]

    def __u_move_clockwise_internal(self, layer):
        layer -= 1
        ORIGINAL_STATE = deepcopy(self._STATE_OF_CUBE)
        for i in range(self._SIZE_OF_CUBE):
            self._STATE_OF_CUBE[1][layer][i] = ORIGINAL_STATE[2][layer][i]
            self._STATE_OF_CUBE[2][layer][i] = ORIGINAL_STATE[3][layer][i]
            self._STATE_OF_CUBE[3][layer][i] = ORIGINAL_STATE[4][layer][i]
            self._STATE_OF_CUBE[4][layer][i] = ORIGINAL_STATE[1][layer][i]

    def __u_move_anticlockwise_internal(self, layer):
        layer -= 1
        ORIGINAL_STATE = deepcopy(self._STATE_OF_CUBE)
        for i in range(self._SIZE_OF_CUBE):
            self._STATE_OF_CUBE[1][layer][i] = ORIGINAL_STATE[4][layer][i]
            self._STATE_OF_CUBE[2][layer][i] = ORIGINAL_STATE[1][layer][i]
            self._STATE_OF_CUBE[3][layer][i] = ORIGINAL_STATE[2][layer][i]
            self._STATE_OF_CUBE[4][layer][i] = ORIGINAL_STATE[3][layer][i]

    def b(self, clockwise=True, layer=1):
        if layer > self._MAX_LAYER or layer <= 0:
            raise Exception("Invalid layer!")
        if clockwise and layer == 1:
            self.__b_move_clockwise_external()
        elif clockwise and layer == self._SIZE_OF_CUBE:
            self.__f_move_anticlockwise_external()
        elif not clockwise and layer == self._SIZE_OF_CUBE:
            self.__f_move_clockwise_external()
        elif clockwise and layer > 1:
            self.__b_move_clockwise_internal(layer=layer)
        elif not clockwise and layer == 1:
            self.__b_move_anticlockwise_external()
        elif not clockwise and layer > 1:
            self.__b_move_anticlockwise_internal(layer=layer)

    def __b_move_clockwise_external(self):
        ORIGINAL_STATE = deepcopy(self._STATE_OF_CUBE)
        self.__clockwise_face_move(4, ORIGINAL_STATE)
        for i in range(self._SIZE_OF_CUBE):
            self._STATE_OF_CUBE[0][0][i] = ORIGINAL_STATE[3][i][-1]
            self._STATE_OF_CUBE[1][i][0] = ORIGINAL_STATE[0][0][self._SIZE_OF_CUBE - i - 1]
            self._STATE_OF_CUBE[3][i][-1] = ORIGINAL_STATE[5][-1][self._SIZE_OF_CUBE - i - 1]
            self._STATE_OF_CUBE[5][-1][i] = ORIGINAL_STATE[1][i][0]

    def __b_move_anticlockwise_external(self):
        ORIGINAL_STATE = deepcopy(self._STATE_OF_CUBE)
        self.__anticlockwise_face_move(4, ORIGINAL_STATE)
        for i in range(self._SIZE_OF_CUBE):
            self._STATE_OF_CUBE[0][0][i] = ORIGINAL_STATE[1][self._SIZE_OF_CUBE - i - 1][0]
            self._STATE_OF_CUBE[1][i][0] = ORIGINAL_STATE[5][-1][i]
            self._STATE_OF_CUBE[3][i][-1] = ORIGINAL_STATE[0][0][i]
            self._STATE_OF_CUBE[5][-1][i] = ORIGINAL_STATE[3][self._SIZE_OF_CUBE - i - 1][-1]

    def __b_move_clockwise_internal(self, layer):
        layer -= 1
        ORIGINAL_STATE = deepcopy(self._STATE_OF_CUBE)
        for i in range(self._SIZE_OF_CUBE):
            self._STATE_OF_CUBE[0][layer][i] = ORIGINAL_STATE[3][i][-1 - layer]
            self._STATE_OF_CUBE[1][i][layer] = ORIGINAL_STATE[0][layer][self._SIZE_OF_CUBE - i - 1]
            self._STATE_OF_CUBE[3][i][-1 - layer] = ORIGINAL_STATE[5][-1 - layer][self._SIZE_OF_CUBE - i - 1]
            self._STATE_OF_CUBE[5][-1 - layer][i] = ORIGINAL_STATE[1][i][layer]

    def __b_move_anticlockwise_internal(self, layer):
        layer -= 1
        ORIGINAL_STATE = deepcopy(self._STATE_OF_CUBE)
        for i in range(self._SIZE_OF_CUBE):
            self._STATE_OF_CUBE[0][layer][i] = ORIGINAL_STATE[1][self._SIZE_OF_CUBE - i - 1][layer]
            self._STATE_OF_CUBE[1][i][layer] = ORIGINAL_STATE[5][-1 - layer][i]
            self._STATE_OF_CUBE[3][i][-1 - layer] = ORIGINAL_STATE[0][layer][i]
            self._STATE_OF_CUBE[5][-1 - layer][i] = ORIGINAL_STATE[3][self._SIZE_OF_CUBE - i - 1][-1 - layer]

    def r(self, clockwise=True, layer=1):
        if layer > self._MAX_LAYER or layer <= 0:
            raise Exception("Invalid layer!")
        if clockwise and layer == 1:
            self.__r_move_clockwise_external()
        elif clockwise and layer == self._SIZE_OF_CUBE:
            self.__l_move_anticlockwise_external()
        elif not clockwise and layer == self._SIZE_OF_CUBE:
            self.__l_move_clockwise_external()
        elif clockwise and layer > 1:
            self.__r_move_clockwise_internal(layer=layer)
        elif not clockwise and layer == 1:
            self.__r_move_anticlockwise_external()
        elif not clockwise and layer > 1:
            self.__r_move_anticlockwise_internal(layer=layer)

    def __r_move_clockwise_external(self):
        ORIGINAL_STATE = deepcopy(self._STATE_OF_CUBE)
        self.__clockwise_face_move(3, ORIGINAL_STATE)
        for i in range(self._SIZE_OF_CUBE):
            self._STATE_OF_CUBE[0][i][-1] = ORIGINAL_STATE[2][i][-1]
            self._STATE_OF_CUBE[2][i][-1] = ORIGINAL_STATE[5][i][-1]
            self._STATE_OF_CUBE[4][i][0] = ORIGINAL_STATE[0][self._SIZE_OF_CUBE - i - 1][-1]
            self._STATE_OF_CUBE[5][i][-1] = ORIGINAL_STATE[4][self._SIZE_OF_CUBE - i - 1][0]

    def __r_move_anticlockwise_external(self):
        ORIGINAL_STATE = deepcopy(self._STATE_OF_CUBE)
        self.__anticlockwise_face_move(3, ORIGINAL_STATE)
        for i in range(self._SIZE_OF_CUBE):
            self._STATE_OF_CUBE[0][i][-1] = ORIGINAL_STATE[4][self._SIZE_OF_CUBE - i - 1][0]
            self._STATE_OF_CUBE[2][i][-1] = ORIGINAL_STATE[0][i][-1]
            self._STATE_OF_CUBE[4][i][0] = ORIGINAL_STATE[5][self._SIZE_OF_CUBE - i - 1][-1]
            self._STATE_OF_CUBE[5][i][-1] = ORIGINAL_STATE[2][i][-1]

    def __r_move_clockwise_internal(self, layer):
        layer -= 1
        ORIGINAL_STATE = deepcopy(self._STATE_OF_CUBE)
        for i in range(self._SIZE_OF_CUBE):
            self._STATE_OF_CUBE[0][i][-1 - layer] = ORIGINAL_STATE[2][i][-1 - layer]
            self._STATE_OF_CUBE[2][i][-1 - layer] = ORIGINAL_STATE[5][i][-1 - layer]
            self._STATE_OF_CUBE[4][i][layer] = ORIGINAL_STATE[0][self._SIZE_OF_CUBE - i - 1][-1 - layer]
            self._STATE_OF_CUBE[5][i][-1 - layer] = ORIGINAL_STATE[4][self._SIZE_OF_CUBE - i - 1][layer]

    def __r_move_anticlockwise_internal(self, layer):
        layer -= 1
        ORIGINAL_STATE = deepcopy(self._STATE_OF_CUBE)
        for i in range(self._SIZE_OF_CUBE):
            self._STATE_OF_CUBE[0][i][-1 - layer] = ORIGINAL_STATE[4][self._SIZE_OF_CUBE - i - 1][layer]
            self._STATE_OF_CUBE[2][i][-1 - layer] = ORIGINAL_STATE[0][i][-1 - layer]
            self._STATE_OF_CUBE[4][i][layer] = ORIGINAL_STATE[5][self._SIZE_OF_CUBE - i - 1][-1 - layer]
            self._STATE_OF_CUBE[5][i][-1 - layer] = ORIGINAL_STATE[2][i][-1 - layer]

    def l(self, clockwise=True, layer=1):
        if layer > self._MAX_LAYER or layer <= 0:
            raise Exception("Invalid layer!")
        if clockwise and layer == 1:
            self.__l_move_clockwise_external()
        elif clockwise and layer == self._SIZE_OF_CUBE:
            self.__r_move_anticlockwise_external()
        elif not clockwise and layer == self._SIZE_OF_CUBE:
            self.__r_move_clockwise_external()
        elif clockwise and layer > 1:
            self.__l_move_clockwise_internal(layer=layer)
        elif not clockwise and layer == 1:
            self.__l_move_anticlockwise_external()
        elif not clockwise and layer > 1:
            self.__l_move_anticlockwise_internal(layer=layer)

    def __l_move_clockwise_external(self):
        ORIGINAL_STATE = deepcopy(self._STATE_OF_CUBE)
        self.__clockwise_face_move(1, ORIGINAL_STATE)
        for i in range(self._SIZE_OF_CUBE):
            self._STATE_OF_CUBE[0][i][0] = ORIGINAL_STATE[4][self._SIZE_OF_CUBE - i - 1][-1]
            self._STATE_OF_CUBE[2][i][0] = ORIGINAL_STATE[0][i][0]
            self._STATE_OF_CUBE[4][i][-1] = ORIGINAL_STATE[5][self._SIZE_OF_CUBE - i - 1][0]
            self._STATE_OF_CUBE[5][i][0] = ORIGINAL_STATE[2][i][0]

    def __l_move_anticlockwise_external(self):
        ORIGINAL_STATE = deepcopy(self._STATE_OF_CUBE)
        self.__anticlockwise_face_move(1, ORIGINAL_STATE)
        for i in range(self._SIZE_OF_CUBE):
            self._STATE_OF_CUBE[0][i][0] = ORIGINAL_STATE[2][i][0]
            self._STATE_OF_CUBE[2][i][0] = ORIGINAL_STATE[5][i][0]
            self._STATE_OF_CUBE[4][i][-1] = ORIGINAL_STATE[0][self._SIZE_OF_CUBE - i - 1][0]
            self._STATE_OF_CUBE[5][i][0] = ORIGINAL_STATE[4][self._SIZE_OF_CUBE - i - 1][-1]

    def __l_move_clockwise_internal(self, layer):
        layer -= 1
        ORIGINAL_STATE = deepcopy(self._STATE_OF_CUBE)
        for i in range(self._SIZE_OF_CUBE):
            self._STATE_OF_CUBE[0][i][layer] = ORIGINAL_STATE[4][self._SIZE_OF_CUBE - i - 1][-1 - layer]
            self._STATE_OF_CUBE[2][i][layer] = ORIGINAL_STATE[0][i][layer]
            self._STATE_OF_CUBE[4][i][-1 - layer] = ORIGINAL_STATE[5][self._SIZE_OF_CUBE - i - 1][layer]
            self._STATE_OF_CUBE[5][i][layer] = ORIGINAL_STATE[2][i][layer]

    def __l_move_anticlockwise_internal(self, layer):
        layer -= 1
        ORIGINAL_STATE = deepcopy(self._STATE_OF_CUBE)
        for i in range(self._SIZE_OF_CUBE):
            self._STATE_OF_CUBE[0][i][layer] = ORIGINAL_STATE[2][i][layer]
            self._STATE_OF_CUBE[2][i][layer] = ORIGINAL_STATE[5][i][layer]
            self._STATE_OF_CUBE[4][i][-1 - layer] = ORIGINAL_STATE[0][self._SIZE_OF_CUBE - i - 1][layer]
            self._STATE_OF_CUBE[5][i][layer] = ORIGINAL_STATE[4][self._SIZE_OF_CUBE - i - 1][-1 - layer]

    def d(self, clockwise=True, layer=1):
        if layer > self._MAX_LAYER or layer <= 0:
            raise Exception("Invalid layer!")
        if clockwise and layer == 1:
            self.__d_move_clockwise_external()
        elif clockwise and layer == self._SIZE_OF_CUBE:
            self.__u_move_anticlockwise_external()
        elif not clockwise and layer == self._SIZE_OF_CUBE:
            self.__u_move_clockwise_external()
        elif clockwise and layer > 1:
            self.__d_move_clockwise_internal(layer=layer)
        elif not clockwise and layer == 1:
            self.__d_move_anticlockwise_external()
        elif not clockwise and layer > 1:
            self.__d_move_anticlockwise_internal(layer=layer)

    def __d_move_clockwise_external(self):
        ORIGINAL_STATE = deepcopy(self._STATE_OF_CUBE)
        self.__clockwise_face_move(5, ORIGINAL_STATE)
        for i in range(self._SIZE_OF_CUBE):
            self._STATE_OF_CUBE[1][-1][i] = ORIGINAL_STATE[4][-1][i]
            self._STATE_OF_CUBE[2][-1][i] = ORIGINAL_STATE[1][-1][i]
            self._STATE_OF_CUBE[3][-1][i] = ORIGINAL_STATE[2][-1][i]
            self._STATE_OF_CUBE[4][-1][i] = ORIGINAL_STATE[3][-1][i]

    def __d_move_anticlockwise_external(self):
        ORIGINAL_STATE = deepcopy(self._STATE_OF_CUBE)
        self.__anticlockwise_face_move(5, ORIGINAL_STATE)
        for i in range(self._SIZE_OF_CUBE):
            self._STATE_OF_CUBE[1][-1][i] = ORIGINAL_STATE[2][-1][i]
            self._STATE_OF_CUBE[2][-1][i] = ORIGINAL_STATE[3][-1][i]
            self._STATE_OF_CUBE[3][-1][i] = ORIGINAL_STATE[4][-1][i]
            self._STATE_OF_CUBE[4][-1][i] = ORIGINAL_STATE[1][-1][i]

    def __d_move_clockwise_internal(self, layer):
        layer -= 1
        ORIGINAL_STATE = deepcopy(self._STATE_OF_CUBE)
        for i in range(self._SIZE_OF_CUBE):
            self._STATE_OF_CUBE[1][-1 - layer][i] = ORIGINAL_STATE[4][-1 - layer][i]
            self._STATE_OF_CUBE[2][-1 - layer][i] = ORIGINAL_STATE[1][-1 - layer][i]
            self._STATE_OF_CUBE[3][-1 - layer][i] = ORIGINAL_STATE[2][-1 - layer][i]
            self._STATE_OF_CUBE[4][-1 - layer][i] = ORIGINAL_STATE[3][-1 - layer][i]

    def __d_move_anticlockwise_internal(self, layer):
        layer -= 1
        ORIGINAL_STATE = deepcopy(self._STATE_OF_CUBE)
        for i in range(self._SIZE_OF_CUBE):
            self._STATE_OF_CUBE[1][-1 - layer][i] = ORIGINAL_STATE[2][-1 - layer][i]
            self._STATE_OF_CUBE[2][-1 - layer][i] = ORIGINAL_STATE[3][-1 - layer][i]
            self._STATE_OF_CUBE[3][-1 - layer][i] = ORIGINAL_STATE[4][-1 - layer][i]
            self._STATE_OF_CUBE[4][-1 - layer][i] = ORIGINAL_STATE[1][-1 - layer][i]

    def show(self, colors=False, icons=True):
        if colors and icons:
            self.__show_with_colors(icons=True)
        elif colors and not icons:
            self.__show_with_colors(icons=False)
        else:
            self.__show_without_colors()

    # we acquire amount of spaces that should be placed before number, by subtracting length of current element from
    # length of maximum element(for better visual look)
    def __print_spaces(self, cube_element):
        return (self._MAX_NUM_LEN - len(str(cube_element))) * " "

    def __generate_top_face(self):
        print((self._SQUARE_LEN + 1) * " " + "+" + "-" * self._SQUARE_LEN + "+")
        for i in range(self._SIZE_OF_CUBE):
            print((self._SQUARE_LEN + 1) * " " + "| " +
                  self.__print_spaces(self._STATE_OF_CUBE[0][i][0]) +
                  str(self._STATE_OF_CUBE[0][i][0]) + "  ", end="")
            for j in range(self._SIZE_OF_CUBE - 2):
                print(self.__print_spaces(self._STATE_OF_CUBE[0][i][j + 1]) +
                      str(self._STATE_OF_CUBE[0][i][j + 1]) + "  ", end="")
            print(self.__print_spaces(self._STATE_OF_CUBE[0][i][-1]) +
                  str(self._STATE_OF_CUBE[0][i][-1]) + " |")

    def __generate_middle_faces(self):
        print("+" + self._SQUARE_LEN * "-" + "+" + self._SQUARE_LEN * "-" + "+" + self._SQUARE_LEN * "-" + "+" +
              self._SQUARE_LEN * "-" + "+")
        for i in range(self._SIZE_OF_CUBE):
            print("| ", end="")
            for j in range(4):
                for k in range(self._SIZE_OF_CUBE - 1):
                    print(self.__print_spaces(self._STATE_OF_CUBE[j + 1][i][k]) +
                          str(self._STATE_OF_CUBE[j + 1][i][k]) + "  ", end="")
                print(self.__print_spaces(self._STATE_OF_CUBE[j + 1][i][-1]) +
                      str(self._STATE_OF_CUBE[j + 1][i][-1]) + " | ", end="")
            print("")
        print("+" + "-" * self._SQUARE_LEN + "+" + "-" * self._SQUARE_LEN + "+" + "-" * self._SQUARE_LEN + "+" +
              "-" * self._SQUARE_LEN + "+")

    def __generate_bottom_face(self):
        for i in range(self._SIZE_OF_CUBE):
            print(" " * (self._SQUARE_LEN + 1) + "| " + self.__print_spaces(self._STATE_OF_CUBE[-1][i][0]) +
                  str(self._STATE_OF_CUBE[-1][i][0]) + "  ", end="")
            for j in range(self._SIZE_OF_CUBE - 2):
                print(self.__print_spaces(self._STATE_OF_CUBE[-1][i][j + 1]) +
                      str(self._STATE_OF_CUBE[-1][i][j + 1]) + "  ", end="")
            print(self.__print_spaces(self._STATE_OF_CUBE[-1][i][-1]) +
                  str(self._STATE_OF_CUBE[-1][i][-1]) + " |")
        print(" " * (self._SQUARE_LEN + 1) + "+" + "-" * self._SQUARE_LEN + "+")

    def __show_without_colors(self):
        self.__generate_top_face()
        self.__generate_middle_faces()
        self.__generate_bottom_face()

    def __show_with_colors(self, icons=True):
        color_icon_config = {
            "empty": "⬛",
            "U": "🟪",
            "L": "🟧",
            "F": "🟩",
            "R": "🟥",
            "B": "🟦",
            "D": "🟨"
        }

        color_text_config = {
            "empty": " ",
            "U": "W",
            "L": "O",
            "F": "G",
            "R": "R",
            "B": "B",
            "D": "Y"
        }
        for i in range(self._SIZE_OF_CUBE):
            for j in range(self._SIZE_OF_CUBE):
                if not icons:
                    print(color_text_config["empty"], end="")
                else:
                    print(color_icon_config["empty"], end="")
            print(" ", end="")
            for j in range(self._SIZE_OF_CUBE):
                face = self._get_face_by_element(self._STATE_OF_CUBE[0][i][j])
                if not icons:
                    print(color_text_config[face], end="")
                else:
                    print(color_icon_config[face], end="")
            print("")
        print("")
        for i in range(self._SIZE_OF_CUBE):
            for j in range(1, 5):
                for k in range(self._SIZE_OF_CUBE):
                    face = self._get_face_by_element(self._STATE_OF_CUBE[j][i][k])
                    if not icons:
                        print(color_text_config[face], end="")
                    else:
                        print(color_icon_config[face], end="")
                print(" ", end="")
            print("")
        print("")
        for i in range(self._SIZE_OF_CUBE):
            for j in range(self._SIZE_OF_CUBE):
                if not icons:
                    print(color_text_config["empty"], end="")
                else:
                    print(color_icon_config["empty"], end="")
            print(" ", end="")
            for j in range(self._SIZE_OF_CUBE):
                face = self._get_face_by_element(self._STATE_OF_CUBE[5][i][j])
                if not icons:
                    print(color_text_config[face], end="")
                else:
                    print(color_icon_config[face], end="")
            print("")
        print("")
        print("")

    def _get_face_by_element(self, element: int):
        square = self._SIZE_OF_CUBE * self._SIZE_OF_CUBE
        if element in self._UP_COLOR_ELEMENTS:
            return "U"
        elif element in self._LEFT_COLOR_ELEMENTS:
            return "L"
        elif element in self._FRONT_COLOR_ELEMENTS:
            return "F"
        elif element in self._RIGHT_COLOR_ELEMENTS:
            return "R"
        elif element in self._BACK_COLOR_ELEMENTS:
            return "B"
        elif element in self._DOWN_COLOR_ELEMENTS:
            return "D"

    def move(self, moves):
        self.__match_moves(moves)

    def __get_layer_len_from_move(self, move):
        n = 0
        for i in move:
            if not i.isdigit():
                break
            n += 1
        return n

    def __match_moves(self, moves):
        moves = moves.strip().split(" ")
        moves_dict = {
            "R": self.r,
            "L": self.l,
            "F": self.f,
            "U": self.u,
            "B": self.b,
            "D": self.d
        }
        ext_move_double = re.compile(r"^[RLFUBD]2$")  # U2, F2, ...
        ext_move_prime = re.compile(r"^[RLFUBD]'$")  # U', F', ...
        ext_move_single = re.compile(r"^[RLFUBD]$")  # U, F, ...
        int_move_single = re.compile(r"^[0-9]+[RLFUBD]$")  # 2U, 3F, ...
        int_move_prime = re.compile(r"^[0-9]+[RLFUBD]'$")  # 2U', 3F', ...
        int_move_double = re.compile(r"^[0-9]+[RLFUBD]2$")  # 2U2, 3F2, ...
        wide_move_single = re.compile(r"^[RLFUBD]w$")  # Uw, Fw, ...
        wide_move_prime = re.compile(r"^[RLFUBD]w'$")  # Uw', Fw', ...
        wide_move_double = re.compile(r"^[RLFUBD]w2$")  # Uw2, Fw2, ...
        adv_wide_move_single = re.compile(r"^[0-9]+[RLFUBD]w$")  # 2Uw, 3Fw, ...
        adv_wide_move_prime = re.compile(r"^[0-9]+[RLFUBD]w'$")  # 2Uw', 3Fw', ...
        adv_wide_move_double = re.compile(r"^[0-9]+[RLFUBD]w2$")  # 2Uw2, 3Fw2, ...
        for move in moves:
            move_index = self.__get_layer_len_from_move(move)  # only for use in moves that start with digit
            # U2, F2, ...
            if re.search(ext_move_double, move):
                moves_dict[move[0]]()
                moves_dict[move[0]]()
            # U', F', ...
            elif re.search(ext_move_prime, move):
                moves_dict[move[0]](clockwise=False)
            # U, F, ...
            elif re.search(ext_move_single, move):
                moves_dict[move]()
            # 2U, 3F, ...
            elif re.search(int_move_single, move):
                moves_dict[move[move_index]](layer=int(move[0]))
            # 2U', 3F', ...
            elif re.search(int_move_prime, move):
                moves_dict[move[move_index]](clockwise=False, layer=int(move[0]))
            # 2U2, 3F2, ...
            elif re.search(int_move_double, move):
                moves_dict[move[move_index]](layer=int(move[0]))
                moves_dict[move[move_index]](layer=int(move[0]))
            # Uw, Fw, ...
            elif re.search(wide_move_single, move):
                moves_dict[move[0]](layer=1)
                moves_dict[move[0]](layer=2)
            # Uw', Fw', ...
            elif re.search(wide_move_prime, move):
                moves_dict[move[0]](clockwise=False, layer=1)
                moves_dict[move[0]](clockwise=False, layer=2)
            # Uw2, Fw2, ...
            elif re.search(wide_move_double, move):
                moves_dict[move[0]](clockwise=False, layer=1)
                moves_dict[move[0]](clockwise=False, layer=2)
                moves_dict[move[0]](clockwise=False, layer=1)
                moves_dict[move[0]](clockwise=False, layer=2)
            # 2Uw, 3Fw, ...
            elif re.search(adv_wide_move_single, move):
                for i in range(int(move[0:move_index])):
                    moves_dict[move[move_index]](layer=i + 1)
            # 2Uw', 3Fw', ...
            elif re.search(adv_wide_move_prime, move):
                for i in range(int(move[0:move_index])):
                    moves_dict[move[move_index]](clockwise=False, layer=i + 1)
            # 2Uw2, 3Fw2, ...
            elif re.search(adv_wide_move_double, move):
                for i in range(int(move[0:move_index])):
                    moves_dict[move[move_index]](layer=i + 1)
                    moves_dict[move[move_index]](layer=i + 1)
            else:
                raise Exception("Invalid move!")

    def scramble(self, number_of_moves=30):
        scramble = ""
        moves = ["R", "L", "U", "D", "B", "F"]
        basic_moves = deepcopy(moves)
        for move in basic_moves:
            for i in range(1, self._SIZE_OF_CUBE // 2):
                moves.append(str(i + 1) + move + "w")
                moves.append(str(i + 1) + move)
        for i in range(number_of_moves):
            random_move = random.choice(moves)
            move_instance = random.randint(1, 3)
            if move_instance == 1:
                scramble += random_move + " "
            elif move_instance == 2:
                scramble += random_move + "' "
            elif move_instance == 3:
                scramble += random_move + "2 "
            else:
                raise Exception("Invalid instance of move!")
        self.__match_moves(scramble)
        return scramble

    def get_state(self):
        return self._STATE_OF_CUBE

    def set_state(self, state):
        self._STATE_OF_CUBE = state

    def define_state(self, cube_string: str):
        self._UP_COLOR_ELEMENTS = []
        self._LEFT_COLOR_ELEMENTS = []
        self._FRONT_COLOR_ELEMENTS = []
        self._RIGHT_COLOR_ELEMENTS = []
        self._BACK_COLOR_ELEMENTS = []
        self._DOWN_COLOR_ELEMENTS = []
        cube_string_index = 0
        if len(cube_string) != self._SIZE_OF_CUBE * self._SIZE_OF_CUBE * 6:
            raise Exception("Invalid number of elements!")
        for face in range(6):
            for row in range(self._SIZE_OF_CUBE):
                for col in range(self._SIZE_OF_CUBE):
                    element = self._STATE_OF_CUBE[face][row][col]
                    if cube_string[cube_string_index] == "W":
                        self._UP_COLOR_ELEMENTS.append(element)
                    elif cube_string[cube_string_index] == "O":
                        self._LEFT_COLOR_ELEMENTS.append(element)
                    elif cube_string[cube_string_index] == "G":
                        self._FRONT_COLOR_ELEMENTS.append(element)
                    elif cube_string[cube_string_index] == "R":
                        self._RIGHT_COLOR_ELEMENTS.append(element)
                    elif cube_string[cube_string_index] == "B":
                        self._BACK_COLOR_ELEMENTS.append(element)
                    elif cube_string[cube_string_index] == "Y":
                        self._DOWN_COLOR_ELEMENTS.append(element)
                    else:
                        raise Exception("Invalid element!")
                    cube_string_index += 1

    def reset(self):
        self._STATE_OF_CUBE = self.__generate_clear_state()

    def __add_plane(self, plotter, center: (float, float, float), direction: (float, float, float), color: str):
        if color is not None:
            plane = pv.Plane(center=(center[0], center[1], center[2]),
                             direction=(direction[0], direction[1], direction[2]),
                             i_size=0.85, j_size=0.85)
            plotter.add_mesh(plane, color=color)

    def __add_number(self, plotter, number: int, center: (float, float, float), direction: (float, float, float),
                     rotate_x=False):
        number_width = 0.7
        if len(str(number)) == 1:
            number_width = 0.3
        if number is not None:
            text = pv.Text3D(str(number), width=number_width, height=0.7, depth=0.0,
                             normal=(direction[0], direction[1], direction[2]),
                             center=(center[0], center[1], center[2]))
            if rotate_x:
                text = text.rotate_x(180)
            plotter.add_mesh(text)

    def __add_cubie(self, plotter, center: (float, float, float),
                    up_color=None, down_color=None,
                    left_color=None, right_color=None,
                    front_color=None, back_color=None,
                    up_number=None, down_number=None,
                    left_number=None, right_number=None,
                    front_number=None, back_number=None):
        x = center[0]
        y = center[1]
        z = center[2]
        plotter.add_mesh(pv.Cube(center=center), color="black")
        self.__add_plane(plotter, center=(x + 0.501, y, z), direction=(1, 0, 0), color=front_color)
        self.__add_number(plotter, number=front_number, center=(x + 0.502, y, z), direction=(1, 0, 0))
        self.__add_plane(plotter, center=(x - 0.501, y, z), direction=(1, 0, 0), color=back_color)
        self.__add_number(plotter, number=back_number, center=(x - 0.502, y, z), direction=(-1, 0, 0), rotate_x=True)
        self.__add_plane(plotter, center=(x, y + 0.501, z), direction=(0, 1, 0), color=right_color)
        self.__add_number(plotter, number=right_number, center=(x, y + 0.502, z), direction=(0, 1, 0))
        self.__add_plane(plotter, center=(x, y - 0.501, z), direction=(0, 1, 0), color=left_color)
        self.__add_number(plotter, number=left_number, center=(x, y - 0.502, z), direction=(0, -1, 0))
        self.__add_plane(plotter, center=(x, y, z + 0.501), direction=(0, 0, 1), color=up_color)
        self.__add_number(plotter, number=up_number, center=(x, y, z + 0.502), direction=(0, 0, 0.5))
        self.__add_plane(plotter, center=(x, y, z - 0.501), direction=(0, 0, 1), color=down_color)
        self.__add_number(plotter, number=down_number, center=(x, y, z - 0.502), direction=(0, 0, -1))




    def show_3d(self, numbers=False, notebook=False, window_size=(1000, 500)):
        color_config = {
            "U": "white",
            "L": "orange",
            "F": "green",
            "R": "red",
            "B": "royalblue",
            "D": "yellow"
        }
        if notebook:
            pv.set_jupyter_backend("html")
            plotter = pv.Plotter(notebook=True, window_size=[window_size[0], window_size[1]])
        else:
            plotter = pv.Plotter()
        center_cube = pv.Cube(x_length=self.CENTER_3DCUBE_SIZE,
                              y_length=self.CENTER_3DCUBE_SIZE,
                              z_length=self.CENTER_3DCUBE_SIZE,)
        plotter.add_mesh(center_cube, color="black")
        self.__plot_3dcube_corners(plotter, color_config, numbers)
        self.__plot_3dcube_y_axis_angles(plotter, color_config, numbers)
        self.__plot_3dcube_x_axis_angles(plotter, color_config, numbers)
        self.__plot_3dcube_z_axis_angles(plotter, color_config, numbers)
        self.__plot_3dcube_x_axis_centers(plotter, color_config, numbers)
        self.__plot_3dcube_y_axis_centers(plotter, color_config, numbers)
        self.__plot_3dcube_z_axis_centers(plotter, color_config, numbers)
        plotter.show()


    def __reset_nums(self):
        self.__up_num = None
        self.__left_num = None
        self.__front_num = None
        self.__right_num = None
        self.__back_num = None
        self.__down_num = None


    def __set_and_reset_nums(self, numbers: bool, up_num=None, left_num=None,
                             front_num=None, right_num=None, back_num=None, down_num=None):
        self.__reset_nums()
        if up_num and numbers:
            self.__up_num = up_num
        if left_num and numbers:
            self.__left_num = left_num
        if front_num and numbers:
            self.__front_num = front_num
        if right_num and numbers:
            self.__right_num = right_num
        if back_num and numbers:
            self.__back_num = back_num
        if down_num and numbers:
            self.__down_num = down_num


    def __plot_3dcube_corners(self, plotter, color_config, numbers: bool):
        half = self.CENTER_3DCUBE_SIZE / 2
        n = self._SIZE_OF_CUBE-1

        self.__set_and_reset_nums(numbers, up_num=self._STATE_OF_CUBE[0][n][0],
                                  left_num=self._STATE_OF_CUBE[1][0][n],
                                  front_num=self._STATE_OF_CUBE[2][0][0])

        self.__add_cubie(plotter, center=(half, -half, half),
                         up_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[0][n][0])],
                         left_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[1][0][n])],
                         front_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[2][0][0])],
                         up_number=self.__up_num,
                         left_number=self.__left_num,
                         front_number=self.__front_num)

        self.__set_and_reset_nums(numbers, up_num=self._STATE_OF_CUBE[0][n][n],
                                  front_num=self._STATE_OF_CUBE[2][0][n],
                                  right_num=self._STATE_OF_CUBE[3][0][0])

        self.__add_cubie(plotter, center=(half, half, half),
                         up_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[0][n][n])],
                         front_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[2][0][n])],
                         right_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[3][0][0])],
                         up_number=self.__up_num,
                         front_number=self.__front_num,
                         right_number=self.__right_num)

        self.__set_and_reset_nums(numbers, left_num=self._STATE_OF_CUBE[1][n][n],
                                  front_num=self._STATE_OF_CUBE[2][n][0],
                                  down_num=self._STATE_OF_CUBE[5][0][0])

        self.__add_cubie(plotter, center=(half, -half, -half),
                         left_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[1][n][n])],
                         front_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[2][n][0])],
                         down_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[5][0][0])],
                         left_number=self.__left_num,
                         front_number=self.__front_num,
                         down_number=self.__down_num)

        self.__set_and_reset_nums(numbers, right_num=self._STATE_OF_CUBE[3][n][0],
                                  front_num=self._STATE_OF_CUBE[2][n][n],
                                  down_num=self._STATE_OF_CUBE[5][0][n])

        self.__add_cubie(plotter, center=(half, half, -half),
                         right_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[3][n][0])],
                         front_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[2][n][n])],
                         down_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[5][0][n])],
                         right_number=self.__right_num,
                         front_number=self.__front_num,
                         down_number=self.__down_num)

        self.__set_and_reset_nums(numbers, left_num=self._STATE_OF_CUBE[1][0][0],
                                  up_num=self._STATE_OF_CUBE[0][0][0],
                                  back_num=self._STATE_OF_CUBE[4][n][0])

        self.__add_cubie(plotter, center=(-half, -half, half),
                         left_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[1][0][0])],
                         up_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[0][0][0])],
                         back_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[4][0][n])],
                         left_number=self.__left_num,
                         up_number=self.__up_num,
                         back_number=self.__back_num)

        self.__set_and_reset_nums(numbers, up_num=self._STATE_OF_CUBE[0][0][n],
                                  right_num=self._STATE_OF_CUBE[3][0][n],
                                  back_num=self._STATE_OF_CUBE[4][n][n])

        self.__add_cubie(plotter, center=(-half, half, half),
                         up_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[0][0][n])],
                         right_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[3][0][n])],
                         back_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[4][0][0])],
                         up_number=self.__up_num,
                         right_number=self.__right_num,
                         back_number=self.__back_num)

        self.__set_and_reset_nums(numbers, right_num=self._STATE_OF_CUBE[3][n][n],
                                  back_num=self._STATE_OF_CUBE[4][0][n],
                                  down_num=self._STATE_OF_CUBE[5][n][n])

        self.__add_cubie(plotter, center=(-half, half, -half),
                         right_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[3][n][n])],
                         back_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[4][n][0])],
                         down_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[5][n][n])],
                         right_number=self.__right_num,
                         back_number=self.__back_num,
                         down_number=self.__down_num)

        self.__set_and_reset_nums(numbers, left_num=self._STATE_OF_CUBE[1][n][0],
                                  down_num=self._STATE_OF_CUBE[5][n][0],
                                  back_num=self._STATE_OF_CUBE[4][0][0])

        self.__add_cubie(plotter, center=(-half, -half, -half),
                         left_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[1][n][0])],
                         down_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[5][n][0])],
                         back_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[4][n][n])],
                         left_number=self.__left_num,
                         down_number=self.__down_num,
                         back_number=self.__back_num)

    def __plot_3dcube_y_axis_angles(self, plotter, color_config, numbers: bool):
        half = self.CENTER_3DCUBE_SIZE / 2
        n = self._SIZE_OF_CUBE - 1
        index = 1
        for y_axis in np.arange(-half + 1 + self.GAP_SIZE, half, 1+self.GAP_SIZE):

            self.__set_and_reset_nums(numbers, front_num=self._STATE_OF_CUBE[2][0][index],
                                      up_num=self._STATE_OF_CUBE[0][n][index])

            self.__add_cubie(plotter, center=(half, y_axis, half),
                             front_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[2][0][index])],
                             up_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[0][n][index])],
                             front_number=self.__front_num,
                             up_number=self.__up_num)

            self.__set_and_reset_nums(numbers, front_num=self._STATE_OF_CUBE[2][n][index],
                                      down_num=self._STATE_OF_CUBE[5][0][index])

            self.__add_cubie(plotter, center=(half, y_axis, -half),
                             front_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[2][n][index])],
                             down_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[5][0][index])],
                             front_number=self.__front_num,
                             down_number=self.__down_num)

            self.__set_and_reset_nums(numbers, up_num=self._STATE_OF_CUBE[0][0][index],
                                      back_num=self._STATE_OF_CUBE[4][n][index])

            self.__add_cubie(plotter, center=(-half, y_axis, half),
                             up_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[0][0][index])],
                             back_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[4][0][n - index])],
                             up_number=self.__up_num,
                             back_number=self.__back_num)

            self.__set_and_reset_nums(numbers, down_num=self._STATE_OF_CUBE[5][n][index],
                                      back_num=self._STATE_OF_CUBE[4][0][index])

            self.__add_cubie(plotter, center=(-half, y_axis, -half),
                             down_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[5][n][index])],
                             back_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[4][n][n - index])],
                             down_number=self.__down_num,
                             back_number=self.__back_num)
            index += 1

    def __plot_3dcube_x_axis_angles(self, plotter, color_config, numbers: bool):
        half = self.CENTER_3DCUBE_SIZE / 2
        n = self._SIZE_OF_CUBE - 1
        index = 1
        for x_axis in np.arange(-half + 1 + self.GAP_SIZE, half, 1 + self.GAP_SIZE):
            self.__set_and_reset_nums(numbers, up_num=self._STATE_OF_CUBE[0][index][n],
                                      right_num=self._STATE_OF_CUBE[3][0][n - index])

            self.__add_cubie(plotter, center=(x_axis, half, half),
                             up_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[0][index][n])],
                             right_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[3][0][n - index])],
                             up_number=self.__up_num, right_number=self.__right_num)

            self.__set_and_reset_nums(numbers, down_num=self._STATE_OF_CUBE[5][n - index][n],
                                      right_num=self._STATE_OF_CUBE[3][n][n - index])

            self.__add_cubie(plotter, center=(x_axis, half, -half),
                             down_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[5][n - index][n])],
                             right_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[3][n][n - index])],
                             down_number=self.__down_num, right_number=self.__right_num)

            self.__set_and_reset_nums(numbers, up_num=self._STATE_OF_CUBE[0][index][0],
                                      left_num=self._STATE_OF_CUBE[1][0][index])

            self.__add_cubie(plotter, center=(x_axis, -half, half),
                             up_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[0][index][0])],
                             left_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[1][0][index])],
                             up_number=self.__up_num, left_number=self.__left_num)

            self.__set_and_reset_nums(numbers, left_num=self._STATE_OF_CUBE[1][n][index],
                                      down_num=self._STATE_OF_CUBE[5][n - index][0])

            self.__add_cubie(plotter, center=(x_axis, -half, -half),
                             left_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[1][n][index])],
                             down_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[5][n - index][0])],
                             left_number=self.__left_num, down_number=self.__down_num)

            index += 1


    def __plot_3dcube_z_axis_angles(self, plotter, color_config, numbers: bool):
        half = self.CENTER_3DCUBE_SIZE / 2
        n = self._SIZE_OF_CUBE - 1
        index = 1
        for z_axis in np.arange(half - 1 - self.GAP_SIZE, -half, -1 - self.GAP_SIZE):

            self.__set_and_reset_nums(numbers, front_num=self._STATE_OF_CUBE[2][index][n],
                                      right_num=self._STATE_OF_CUBE[3][index][0])

            self.__add_cubie(plotter, center=(half, half, z_axis),
                             front_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[2][index][n])],
                             right_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[3][index][0])],
                             front_number=self.__front_num, right_number=self.__right_num)

            self.__set_and_reset_nums(numbers, front_num=self._STATE_OF_CUBE[2][index][0],
                                      left_num=self._STATE_OF_CUBE[1][index][n])

            self.__add_cubie(plotter, center=(half, -half, z_axis),
                             front_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[2][index][0])],
                             left_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[1][index][n])],
                             front_number=self.__front_num, left_number=self.__left_num)

            self.__set_and_reset_nums(numbers, right_num=self._STATE_OF_CUBE[3][index][n],
                                      back_num=self._STATE_OF_CUBE[4][n-index][n])

            self.__add_cubie(plotter, center=(-half, half, z_axis),
                             right_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[3][index][n])],
                             back_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[4][index][0])],
                             right_number=self.__right_num, back_number=self.__back_num)

            self.__set_and_reset_nums(numbers, back_num=self._STATE_OF_CUBE[4][n-index][0],
                                      left_num=self._STATE_OF_CUBE[1][index][0])

            self.__add_cubie(plotter, center=(-half, -half, z_axis),
                             back_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[4][index][n])],
                             left_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[1][index][0])],
                             back_number=self.__back_num, left_number=self.__left_num)

            index += 1


    def __plot_3dcube_x_axis_centers(self, plotter, color_config, numbers: bool):
        half = self.CENTER_3DCUBE_SIZE / 2
        n = self._SIZE_OF_CUBE - 1
        row = 1
        for z_axis in np.arange(half - 1 - self.GAP_SIZE, -half, -1 - self.GAP_SIZE):
            col = 1
            for y_axis in np.arange(-half + 1 + self.GAP_SIZE, half, 1 + self.GAP_SIZE):

                self.__set_and_reset_nums(numbers, front_num=self._STATE_OF_CUBE[2][row][col])

                self.__add_cubie(plotter, center=(half, y_axis, z_axis),
                                 front_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[2][row][col])],
                                 front_number=self.__front_num)

                self.__set_and_reset_nums(numbers, back_num=self._STATE_OF_CUBE[4][n-row][col])

                self.__add_cubie(plotter, center=(-half, y_axis, z_axis),
                                 back_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[4][row][n - col])],
                                 back_number=self.__back_num)

                col += 1
            row += 1

    def __plot_3dcube_y_axis_centers(self, plotter, color_config, numbers: bool):
        half = self.CENTER_3DCUBE_SIZE / 2
        n = self._SIZE_OF_CUBE - 1
        row = 1
        for z_axis in np.arange(half - 1 - self.GAP_SIZE, -half, -1 - self.GAP_SIZE):
            col = 1
            for x_axis in np.arange(-half + 1 + self.GAP_SIZE, half, 1 + self.GAP_SIZE):

                self.__set_and_reset_nums(numbers, right_num=self._STATE_OF_CUBE[3][row][n - col])

                self.__add_cubie(plotter, center=(x_axis, half, z_axis),
                                 right_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[3][row][n - col])],
                                 right_number=self.__right_num)

                self.__set_and_reset_nums(numbers, left_num=self._STATE_OF_CUBE[1][row][col])

                self.__add_cubie(plotter, center=(x_axis, -half, z_axis),
                                 left_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[1][row][col])],
                                 left_number=self.__left_num)

                col += 1
            row += 1


    def __plot_3dcube_z_axis_centers(self, plotter, color_config, numbers: bool):
        half = self.CENTER_3DCUBE_SIZE / 2
        n = self._SIZE_OF_CUBE - 1
        row = 1
        for x_axis in np.arange(-half + 1 + self.GAP_SIZE, half, 1 + self.GAP_SIZE):
            col = 1
            for y_axis in np.arange(-half + 1 + self.GAP_SIZE, half, 1 + self.GAP_SIZE):

                self.__set_and_reset_nums(numbers, up_num=self._STATE_OF_CUBE[0][row][col])

                self.__add_cubie(plotter, center=(x_axis, y_axis, half),
                                 up_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[0][row][col])],
                                 up_number=self.__up_num)

                self.__set_and_reset_nums(numbers, down_num=self._STATE_OF_CUBE[5][n - row][col])

                self.__add_cubie(plotter, center=(x_axis, y_axis, -half),
                                 down_color=color_config[self._get_face_by_element(self._STATE_OF_CUBE[5][n - row][col])],
                                 down_number=self.__down_num)

                col += 1
            row += 1





# ########################################  Miscellaneous functions ####################################################

def rotate_right(arr, l, r):
    """"Rotates array arr right between l and r. r is included."""
    temp = arr[r]
    for i in range(r, l, -1):
        arr[i] = arr[i-1]
    arr[l] = temp


def rotate_left(arr, l, r):
    """"Rotates array arr left between l and r. r is included."""
    temp = arr[l]
    for i in range(l, r):
        arr[i] = arr[i+1]
    arr[r] = temp


# ####### The cube on the coordinate level is described by a 3-tuple of natural numbers in phase 1 and phase 2. ########

from supercube.kociemba_2x2 import moves as mv
from supercube.kociemba_2x2.defs import N_MOVE

SOLVED = 0  # 0 is index of solved state (except for u_edges coordinate)


class CoordCube:
    """Represents a 2x2x2 cube on the coordinate level.

    A state is uniquely determined by the coordinates corntwist and cornperm.
    """

    def __init__(self, cc=None):
        if cc is None:
            self.corntwist = SOLVED  # twist of corners
            self.cornperm = SOLVED  # permutation of corners
        else:
            self.corntwist = cc.get_corntwist()
            self.cornperm = cc.get_cornperm()

    def __str__(self):
        s = '(corntwist: ' + str(self.corntwist) + ', cornperm: ' + str(self.cornperm) + ')'
        return s

    def move(self, m):
        self.corntwist = mv.twist_move[N_MOVE * self.corntwist + m]
        self.cornperm = mv.corners_move[N_MOVE * self.cornperm + m]
########################################################################################################################




from setuptools import setup, find_packages

DESCRIPTION="Library for simulating rubik's cubes of different sizes (N x N), for solving 2 x 2 and 3 x 3 cubes with kociemba algorithm, and for 3D visualization of cubes."
LONG_DESCRIPTION = open("README.md", encoding="utf8").read()
project_urls = {
    "Homepage": "https://github.com/Ilmi28/SuperCube"
}


setup(
    name="super-rubik-cube",
    version="1.0.53",
    author="Ilmi28",
    author_email="<ilmialiev28@gmail.com>",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=LONG_DESCRIPTION,
    package_data={'supercube': ['kociemba/pykociemba/prunetables/*']},
    packages=find_packages(),
    keywords=['python', 'rubik', 'cube', 'solve', '3x3', '2x2', "rubik's cube", 'pocket cube', 'NxN', 'kociemba', 'supercube', 'solver', '3D', 'jupyter'],
    install_requires=['pyvista[jupyter]==0.43.10', 'numpy<2', 'pythreejs', 'ipywidgets'],
    project_urls=project_urls,
    license='GPL-3.0'
)



from supercube import kociemba
from .pykociemba import search

def _solve(cube, pattern, max_depth=24):
    errors = {
        'Error 1': 'There is not exactly one facelet of each colour',
        'Error 2': 'Not all 12 edges exist exactly once',
        'Error 3': 'Flip error: One edge has to be flipped',
        'Error 4': 'Not all corners exist exactly once',
        'Error 5': 'Twist error: One corner has to be twisted',
        'Error 6': 'Parity error: Two corners or two edges have to be exchanged',
        'Error 7': 'No solution exists for the given maxDepth',
        'Error 8': 'Timeout, no solution within given time'
    }
    if pattern is not None:
        cube = search.patternize(cube, pattern)
    res = search.Search().solution(cube, max_depth, 1000, False).strip()
    if res in errors:
        raise ValueError(errors[res])
    else:
        return res

def solve(cubestring, patternstring=None, max_depth=24):
    """
    Solve a Rubik's cube using two-phase algorithm.

    >>> solve("BBURUDBFUFFFRRFUUFLULUFUDLRRDBBDBDBLUDDFLLRRBRLLLBRDDF")
    "B U' L' D' R' D' L2 D' L F' L' D F2 R2 U R2 B2 U2 L2 F2 D'"

    >>> kociemba.solve('FLBUULFFLFDURRDBUBUUDDFFBRDDBLRDRFLLRLRULFUDRRBDBBBUFL', 'BBURUDBFUFFFRRFUUFLULUFUDLRRDBBDBDBLUDDFLLRRBRLLLBRDDF')
    u"R' D2 R' U2 R F2 D B2 U' R F' U R2 D L2 D' B2 R2 B2 U' B2"
    """

    return _solve(cubestring, patternstring, max_depth)

__all__ = ['solve']



import time
from builtins import range
from .color import colors
from .facecube import FaceCube
from .coordcube import CoordCube, getPruning
from .cubiecube import CubieCube

class Search(object):
    """Class Search implements the Two-Phase-Algorithm."""

    ax_to_s = ["U", "R", "F", "D", "L", "B"]
    po_to_s = [None, " ", "2 ", "' "]

    def __init__(self):
        self.ax              = [0] * 31  # The axis of the move
        self.po              = [0] * 31  # The power of the move
        self.flip            = [0] * 31  # phase1 coordinates
        self.twist           = [0] * 31
        self.slice           = [0] * 31
        self.parity          = [0] * 31  # phase2 coordinates
        self.URFtoDLF        = [0] * 31
        self.FRtoBR          = [0] * 31
        self.URtoUL          = [0] * 31
        self.UBtoDF          = [0] * 31
        self.URtoDF          = [0] * 31
        self.minDistPhase1   = [0] * 31  # IDA* distance do goal estimations
        self.minDistPhase2   = [0] * 31

    def solutionToString(self, length, depthPhase1=None):
        """generate the solution string from the array data"""

        s = ""
        for i in range(length):
            s += self.ax_to_s[self.ax[i]]
            s += self.po_to_s[self.po[i]]
            if depthPhase1 is not None and i == depthPhase1 - 1:
                s += ". "
        return s

    def solution(self, facelets, maxDepth, timeOut, useSeparator):
        """
        Computes the solver string for a given cube.

        @param facelets
                 is the cube definition string, see {@link Facelet} for the format.

        @param maxDepth
                 defines the maximal allowed maneuver length. For random cubes, a maxDepth of 21 usually will return a
                 solution in less than 0.5 seconds. With a maxDepth of 20 it takes a few seconds on average to find a
                 solution, but it may take much longer for specific cubes.

        @param timeOut
                 defines the maximum computing time of the method in seconds. If it does not return with a solution, it returns with
                 an error code.

        @param useSeparator
                 determines if a " . " separates the phase1 and phase2 parts of the solver string like in F' R B R L2 F .
                 U2 U D for example.<br>
        @return The solution string or an error code:<br>
                Error 1: There is not exactly one facelet of each colour<br>
                Error 2: Not all 12 edges exist exactly once<br>
                Error 3: Flip error: One edge has to be flipped<br>
                Error 4: Not all corners exist exactly once<br>
                Error 5: Twist error: One corner has to be twisted<br>
                Error 6: Parity error: Two corners or two edges have to be exchanged<br>
                Error 7: No solution exists for the given maxDepth<br>
                Error 8: Timeout, no solution within given time
        """

        # +++++++++++++++++++++check for wrong input +++++++++++++++++++++++++++++
        count = [0] * 6
        try:
            for i in range(54):
                assert facelets[i] in colors
                count[colors[facelets[i]]] += 1
        except Exception:
            return "Error 1"

        for i in range(6):
            if count[i] != 9:
                return "Error 1"

        fc = FaceCube(facelets)
        cc = fc.toCubieCube()
        s = cc.verify()
        if s != 0:
            return "Error %s" % abs(s)

        # +++++++++++++++++++++++ initialization +++++++++++++++++++++++++++++++++
        c = CoordCube(cc)

        self.po[0] = 0
        self.ax[0] = 0
        self.flip[0] = c.flip
        self.twist[0] = c.twist
        self.parity[0] = c.parity
        self.slice[0] = c.FRtoBR // 24
        self.URFtoDLF[0] = c.URFtoDLF
        self.FRtoBR[0] = c.FRtoBR
        self.URtoUL[0] = c.URtoUL
        self.UBtoDF[0] = c.UBtoDF

        self.minDistPhase1[1] = 1   # else failure for depth=1, n=0
        mv = 0
        n = 0
        busy = False
        depthPhase1 = 1

        tStart = time.time()

        # +++++++++++++++++++ Main loop ++++++++++++++++++++++++++++++++++++++++++
        while True:
            while True:
                if depthPhase1 - n > self.minDistPhase1[n + 1] and not busy:
                    if self.ax[n] == 0 or self.ax[n] == 3:   # Initialize next move
                        n += 1
                        self.ax[n] = 1
                    else:
                        n += 1
                        self.ax[n] = 0
                    self.po[n] = 1
                else:
                    self.po[n] += 1
                    if self.po[n] > 3:
                        while True:
                            # increment axis
                            self.ax[n] += 1
                            if self.ax[n] > 5:

                                if time.time() - tStart > timeOut:
                                    return "Error 8"

                                if n == 0:
                                    if depthPhase1 >= maxDepth:
                                        return "Error 7"
                                    else:
                                        depthPhase1 += 1
                                        self.ax[n] = 0
                                        self.po[n] = 1
                                        busy = False
                                        break
                                else:
                                    n -= 1
                                    busy = True
                                    break

                            else:
                                self.po[n] = 1
                                busy = False

                            if not (n != 0 and (self.ax[n - 1] == self.ax[n] or self.ax[n - 1] - 3 == self.ax[n])):
                                break
                    else:
                        busy = False
                if not busy:
                    break

            # +++++++++++++ compute new coordinates and new minDistPhase1 ++++++++++
            # if minDistPhase1 =0, the H subgroup is reached
            mv = 3 * self.ax[n] + self.po[n] - 1
            self.flip[n + 1] = CoordCube.flipMove[self.flip[n]][mv]
            self.twist[n + 1] = CoordCube.twistMove[self.twist[n]][mv]
            self.slice[n + 1] = CoordCube.FRtoBR_Move[self.slice[n] * 24][mv] // 24
            self.minDistPhase1[n + 1] = max(
                getPruning(
                    CoordCube.Slice_Flip_Prun,
                    CoordCube.N_SLICE1 * self.flip[n + 1] + self.slice[n + 1]
                ),
                getPruning(
                    CoordCube.Slice_Twist_Prun,
                    CoordCube.N_SLICE1 * self.twist[n + 1] + self.slice[n + 1]
                )
            )
            # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

            if self.minDistPhase1[n + 1] == 0 and n >= depthPhase1 - 5:
                self.minDistPhase1[n + 1] = 10  # instead of 10 any value >5 is possible
                if n == depthPhase1 - 1:
                    s = self.totalDepth(depthPhase1, maxDepth)
                    if s >= 0:
                        if (s == depthPhase1
                            or (
                                self.ax[depthPhase1 - 1] != self.ax[depthPhase1]
                                and self.ax[depthPhase1 - 1] != self.ax[depthPhase1] + 3)):
                            return self.solutionToString(s, depthPhase1) if useSeparator else self.solutionToString(s)

    def totalDepth(self, depthPhase1, maxDepth):
        """
        Apply phase2 of algorithm and return the combined phase1 and phase2 depth. In phase2, only the moves
        U,D,R2,F2,L2 and B2 are allowed.
        """

        mv = 0
        d1 = 0
        d2 = 0
        maxDepthPhase2 = min(10, maxDepth - depthPhase1)    # Allow only max 10 moves in phase2
        for i in range(depthPhase1):
            mv = 3 * self.ax[i] + self.po[i] - 1
            self.URFtoDLF[i + 1] = CoordCube.URFtoDLF_Move[self.URFtoDLF[i]][mv]
            self.FRtoBR[i + 1] = CoordCube.FRtoBR_Move[self.FRtoBR[i]][mv]
            self.parity[i + 1] = CoordCube.parityMove[self.parity[i]][mv]

        d1 = getPruning(
            CoordCube.Slice_URFtoDLF_Parity_Prun,
            (CoordCube.N_SLICE2 * self.URFtoDLF[depthPhase1] + self.FRtoBR[depthPhase1]) * 2 + self.parity[depthPhase1]
        )
        if d1 > maxDepthPhase2:
            return -1

        for i in range(depthPhase1):
            mv = 3 * self.ax[i] + self.po[i] - 1
            self.URtoUL[i + 1] = CoordCube.URtoUL_Move[self.URtoUL[i]][mv]
            self.UBtoDF[i + 1] = CoordCube.UBtoDF_Move[self.UBtoDF[i]][mv]

        self.URtoDF[depthPhase1] = CoordCube.MergeURtoULandUBtoDF[self.URtoUL[depthPhase1]][self.UBtoDF[depthPhase1]]

        d2 = getPruning(
            CoordCube.Slice_URtoDF_Parity_Prun,
            (CoordCube.N_SLICE2 * self.URtoDF[depthPhase1] + self.FRtoBR[depthPhase1]) * 2 + self.parity[depthPhase1]
        )
        if d2 > maxDepthPhase2:
            return -1

        self.minDistPhase2[depthPhase1] = max(d1, d2)
        if self.minDistPhase2[depthPhase1] == 0:    # already solved
            return depthPhase1

        # now set up search

        depthPhase2 = 1
        n = depthPhase1
        busy = False
        self.po[depthPhase1] = 0
        self.ax[depthPhase1] = 0
        self.minDistPhase2[n + 1] = 1   # else failure for depthPhase2=1, n=0
        # +++++++++++++++++++ end initialization +++++++++++++++++++++++++++++++++

        while True:
            while True:
                if depthPhase1 + depthPhase2 - n > self.minDistPhase2[n + 1] and not busy:

                    if self.ax[n] == 0 or self.ax[n] == 3:    # Initialize next move
                        n += 1
                        self.ax[n] = 1
                        self.po[n] = 2
                    else:
                        n += 1
                        self.ax[n] = 0
                        self.po[n] = 1
                else:
                    if self.ax[n] == 0 or self.ax[n] == 3:
                        self.po[n] += 1
                        _ = (self.po[n] > 3)
                    else:
                        self.po[n] += 2
                        _ = (self.po[n] > 3)
                    if _:
                        while True:
                            # increment axis
                            self.ax[n] += 1
                            if self.ax[n] > 5:
                                if n == depthPhase1:
                                    if depthPhase2 >= maxDepthPhase2:
                                        return -1
                                    else:
                                        depthPhase2 += 1
                                        self.ax[n] = 0
                                        self.po[n] = 1
                                        busy = False
                                        break
                                else:
                                    n -= 1
                                    busy = True
                                    break
                            else:
                                if self.ax[n] == 0 or self.ax[n] == 3:
                                    self.po[n] = 1
                                else:
                                    self.po[n] = 2
                                busy = False

                            if not (n != depthPhase1 and (self.ax[n - 1] == self.ax[n] or self.ax[n - 1] - 3 == self.ax[n])):
                                break

                    else:
                        busy = False

                if not busy:
                    break

            # +++++++++++++ compute new coordinates and new minDist ++++++++++
            mv = 3 * self.ax[n] + self.po[n] - 1

            self.URFtoDLF[n + 1] = CoordCube.URFtoDLF_Move[self.URFtoDLF[n]][mv]
            self.FRtoBR[n + 1] = CoordCube.FRtoBR_Move[self.FRtoBR[n]][mv]
            self.parity[n + 1] = CoordCube.parityMove[self.parity[n]][mv]
            self.URtoDF[n + 1] = CoordCube.URtoDF_Move[self.URtoDF[n]][mv]

            self.minDistPhase2[n + 1] = max(
                getPruning(
                    CoordCube.Slice_URtoDF_Parity_Prun,
                    (CoordCube.N_SLICE2 * self.URtoDF[n + 1] + self.FRtoBR[n + 1]) * 2 + self.parity[n + 1]
                ),
                getPruning(
                    CoordCube.Slice_URFtoDLF_Parity_Prun,
                    (CoordCube.N_SLICE2 * self.URFtoDLF[n + 1] + self.FRtoBR[n + 1]) * 2 + self.parity[n + 1]
                )
            )
            # ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

            if self.minDistPhase2[n + 1] == 0:
                break

        return depthPhase1 + depthPhase2

def patternize(facelets, pattern):
    facelets_cc = FaceCube(facelets).toCubieCube()
    patternized_cc = CubieCube()
    FaceCube(pattern).toCubieCube().invCubieCube(patternized_cc)
    patternized_cc.multiply(facelets_cc)
    return patternized_cc.toFaceCube().to_String()
